@@ -67,7 +67,7 @@ Street, Fifth Floor, Boston, MA 02110-1335 USA
 #include "backup_debug.h"
 
 char *tool_name;
-char tool_args[8192];
+char tool_args[2048];
 
 ulong mysql_server_version;
 
@@ -77,7 +77,6 @@ bool have_lock_wait_timeout = false;
 bool have_galera_enabled = false;
 bool have_multi_threaded_slave = false;
 bool have_gtid_slave = false;
-bool innobase_data_file_path_allocated= false;
 
 /* Kill long selects */
 static mysql_mutex_t kill_query_thread_mutex;
@@ -458,12 +457,9 @@ bool get_mysql_vars(MYSQL *connection)
 
   msg("Using server version %s", version_var);
 
-  if (opt_galera_info && !have_galera_enabled)
+  if (!(ret= detect_mysql_capabilities_for_backup()))
   {
-    msg("--galera-info is specified on the command "
-        "line, but the server does not support Galera "
-        "replication. Ignoring the option.");
-    opt_galera_info= false;
+    goto out;
   }
 
   /* make sure datadir value is the same in configuration file */
@@ -499,19 +495,21 @@ bool get_mysql_vars(MYSQL *connection)
   }
 
   if (innodb_data_file_path_var && *innodb_data_file_path_var)
-    innobase_data_file_path= my_once_strdup(innodb_data_file_path_var,
-                                            MYF(MY_FAE));
+    innobase_data_file_path= my_strdup(PSI_NOT_INSTRUMENTED,
+                                       innodb_data_file_path_var, MYF(MY_FAE));
 
   if (innodb_data_home_dir_var)
-    innobase_data_home_dir= my_once_strdup(innodb_data_home_dir_var,
-                                           MYF(MY_FAE));
+    innobase_data_home_dir= my_strdup(PSI_NOT_INSTRUMENTED,
+                                      innodb_data_home_dir_var, MYF(MY_FAE));
 
   if (innodb_log_group_home_dir_var && *innodb_log_group_home_dir_var)
-    srv_log_group_home_dir= my_once_strdup(innodb_log_group_home_dir_var,
-                                           MYF(MY_FAE));
+    srv_log_group_home_dir= my_strdup(PSI_NOT_INSTRUMENTED,
+                                      innodb_log_group_home_dir_var,
+                                      MYF(MY_FAE));
 
   if (innodb_undo_directory_var && *innodb_undo_directory_var)
-    srv_undo_dir= my_once_strdup(innodb_undo_directory_var, MYF(MY_FAE));
+    srv_undo_dir= my_strdup(PSI_NOT_INSTRUMENTED, innodb_undo_directory_var,
+                            MYF(MY_FAE));
 
   if (innodb_log_file_size_var)
   {
@@ -533,7 +531,10 @@ bool get_mysql_vars(MYSQL *connection)
   }
 
   if (aria_log_dir_path_var)
-    aria_log_dir_path= my_once_strdup(aria_log_dir_path_var, MYF(MY_FAE));
+  {
+    aria_log_dir_path= my_strdup(PSI_NOT_INSTRUMENTED,
+                                 aria_log_dir_path_var, MYF(MY_FAE));
+  }
 
   if (page_zip_level_var != NULL)
   {
@@ -546,10 +547,34 @@ bool get_mysql_vars(MYSQL *connection)
     xb_load_list_string(ignore_db_dirs, ",", register_ignore_db_dirs_filter);
 
 out:
+  free_mysql_variables(mysql_vars);
 
   return (ret);
 }
 
+/*********************************************************************//**
+Query the server to find out what backup capabilities it supports.
+@return	true on success. */
+bool
+detect_mysql_capabilities_for_backup()
+{
+	/* do some sanity checks */
+	if (opt_galera_info && !have_galera_enabled) {
+		msg("--galera-info is specified on the command "
+		 	"line, but the server does not support Galera "
+		 	"replication. Ignoring the option.");
+		opt_galera_info = false;
+	}
+
+	if (opt_slave_info && have_multi_threaded_slave &&
+	    !have_gtid_slave) {
+	    	msg("The --slave-info option requires GTID enabled for a "
+			"multi-threaded slave.");
+		return(false);
+	}
+
+	return(true);
+}
 
 static
 bool
@@ -557,36 +582,12 @@ select_incremental_lsn_from_history(lsn_t *incremental_lsn)
 {
 	MYSQL_RES *mysql_result;
 	char query[1000];
-	char buf[NAME_LEN*2+3];
-
-	size_t opt_incremental_history_name_len= 0;
-	size_t opt_incremental_history_uuid_len= 0;
-
-	if (opt_incremental_history_name)
-		opt_incremental_history_name_len=
-                  strlen(opt_incremental_history_name);
-
-	if (opt_incremental_history_uuid)
-		opt_incremental_history_uuid_len=
-                  strlen(opt_incremental_history_uuid);
-
-	if (opt_incremental_history_name_len*2 > sizeof(buf))
-		die("Incremental history table name '%s' is too long.",
-		    opt_incremental_history_name);
-
-	if (opt_incremental_history_uuid_len*2 > sizeof(buf))
-		die("Incremental history uuid '%s' is too long.",
-		    opt_incremental_history_uuid);
-
-	if (opt_incremental_history_name && opt_incremental_history_name[0]
-	    && opt_incremental_history_uuid && opt_incremental_history_uuid[0])
-		die("It is allowed to use either --incremental-history-name "
-		    "or --incremental-history-uuid, but not both.");
+	char buf[100];
 
 	if (opt_incremental_history_name) {
 		mysql_real_escape_string(mysql_connection, buf,
 				opt_incremental_history_name,
-				(unsigned long) opt_incremental_history_name_len);
+				(unsigned long)strlen(opt_incremental_history_name));
 		snprintf(query, sizeof(query),
 			"SELECT innodb_to_lsn "
 			"FROM " XB_HISTORY_TABLE " "
@@ -595,10 +596,11 @@ select_incremental_lsn_from_history(lsn_t *incremental_lsn)
 			"ORDER BY innodb_to_lsn DESC LIMIT 1",
 			buf);
 	}
-	else if (opt_incremental_history_uuid) {
+
+	if (opt_incremental_history_uuid) {
 		mysql_real_escape_string(mysql_connection, buf,
 				opt_incremental_history_uuid,
-				(unsigned long) opt_incremental_history_uuid_len);
+				(unsigned long)strlen(opt_incremental_history_uuid));
 		snprintf(query, sizeof(query),
 			"SELECT innodb_to_lsn "
 			"FROM " XB_HISTORY_TABLE " "
@@ -608,8 +610,6 @@ select_incremental_lsn_from_history(lsn_t *incremental_lsn)
 			buf);
 	}
 
-	/* xb_mysql_query(..,.., true) will die on error, so
-	mysql_result can't be nullptr */
 	mysql_result = xb_mysql_query(mysql_connection, query, true);
 
 	ut_ad(mysql_num_fields(mysql_result) == 1);
@@ -926,7 +926,7 @@ lock_for_backup_stage_flush(MYSQL *connection) {
 	if (opt_kill_long_queries_timeout) {
 		start_query_killer();
 	}
-	xb_mysql_query(connection, "BACKUP STAGE FLUSH", false);
+	xb_mysql_query(connection, "BACKUP STAGE FLUSH", true);
 	if (opt_kill_long_queries_timeout) {
 		stop_query_killer();
 	}
@@ -938,7 +938,7 @@ lock_for_backup_stage_block_ddl(MYSQL *connection) {
 	if (opt_kill_long_queries_timeout) {
 		start_query_killer();
 	}
-	xb_mysql_query(connection, "BACKUP STAGE BLOCK_DDL", false);
+	xb_mysql_query(connection, "BACKUP STAGE BLOCK_DDL", true);
 	DBUG_MARIABACKUP_EVENT("after_backup_stage_block_ddl", {});
 	if (opt_kill_long_queries_timeout) {
 		stop_query_killer();
@@ -951,7 +951,7 @@ lock_for_backup_stage_commit(MYSQL *connection) {
 	if (opt_kill_long_queries_timeout) {
 		start_query_killer();
 	}
-	xb_mysql_query(connection, "BACKUP STAGE BLOCK_COMMIT", false);
+	xb_mysql_query(connection, "BACKUP STAGE BLOCK_COMMIT", true);
 	DBUG_MARIABACKUP_EVENT("after_backup_stage_block_commit", {});
 	if (opt_kill_long_queries_timeout) {
 		stop_query_killer();
@@ -962,12 +962,12 @@ lock_for_backup_stage_commit(MYSQL *connection) {
 bool backup_lock(MYSQL *con, const char *table_name) {
 	static const std::string backup_lock_prefix("BACKUP LOCK ");
 	std::string backup_lock_query = backup_lock_prefix + table_name;
-	xb_mysql_query(con, backup_lock_query.c_str(), false);
+	xb_mysql_query(con, backup_lock_query.c_str(), true);
 	return true;
 }
 
 bool backup_unlock(MYSQL *con) {
-	xb_mysql_query(con, "BACKUP UNLOCK", false);
+	xb_mysql_query(con, "BACKUP UNLOCK", true);
 	return true;
 }
 
@@ -981,8 +981,6 @@ get_tables_in_use(MYSQL *con) {
 		msg("Table %s is in use", tk.c_str());
 		result.insert(std::move(tk));
 	}
-        if (q_res)
-          mysql_free_result(q_res);
 	return result;
 }
 
@@ -1712,7 +1710,7 @@ write_xtrabackup_info(ds_ctxt *datasink,
 	char buf_end_time[100];
 	tm tm;
 	std::ostringstream oss;
-	const char *xb_stream_name[] = {"file", "mbstream"};
+	const char *xb_stream_name[] = {"file", "tar", "xbstream"};
 
 	uuid = read_mysql_one_value(connection, "SELECT UUID()");
 	server_version = read_mysql_one_value(connection, "SELECT VERSION()");
@@ -1795,10 +1793,6 @@ write_xtrabackup_info(ds_ctxt *datasink,
 		goto cleanup;
 	}
 
-	xb_mysql_query(connection,
-		"ALTER TABLE IF EXISTS " XB_HISTORY_TABLE
-		" MODIFY format ENUM('file', 'tar', 'mbstream') DEFAULT NULL", false);
-
 	xb_mysql_query(connection,
 		"CREATE TABLE IF NOT EXISTS " XB_HISTORY_TABLE "("
 		"uuid VARCHAR(40) NOT NULL PRIMARY KEY,"
@@ -1816,7 +1810,7 @@ write_xtrabackup_info(ds_ctxt *datasink,
 		"innodb_to_lsn BIGINT UNSIGNED DEFAULT NULL,"
 		"partial ENUM('Y', 'N') DEFAULT NULL,"
 		"incremental ENUM('Y', 'N') DEFAULT NULL,"
-		"format ENUM('file', 'tar', 'mbstream') DEFAULT NULL,"
+		"format ENUM('file', 'tar', 'xbstream') DEFAULT NULL,"
 		"compressed ENUM('Y', 'N') DEFAULT NULL"
 		") CHARACTER SET utf8 ENGINE=innodb", false);
 
@@ -1920,7 +1914,7 @@ bool write_backup_config_file(ds_ctxt *datasink)
 		srv_log_file_size,
 		srv_page_size,
 		srv_undo_dir,
-		srv_undo_tablespaces,
+                (uint) srv_undo_tablespaces,
 		page_zip_level,
 		innobase_buffer_pool_filename ?
 			"innodb_buffer_pool_filename=" : "",
@@ -1942,22 +1936,12 @@ char *make_argv(char *buf, size_t len, int argc, char **argv)
 	while (argc > 0 && left > 0)
 	{
 		arg = *argv;
-		if (strncmp(*argv, STRING_WITH_LEN("--password=")) == 0) {
+		if (strncmp(*argv, "--password", strlen("--password")) == 0) {
 			arg = "--password=...";
-		} else
-		if (strcmp(*argv, "--password") == 0) {
-			arg = "--password ...";
-                        ++argv; --argc;
-                } else
-                if (strncmp(*argv, STRING_WITH_LEN("-p")) == 0) {
-			arg = "-p...";
 		}
-
-		uint l= snprintf(buf + len - left, left,
+		left-= snprintf(buf + len - left, left,
 				"%s%c", arg, argc > 1 ? ' ' : 0);
 		++argv; --argc;
-                if (l < left)
-                  left-= l;
 	}
 
 	return buf;
@@ -1967,7 +1951,7 @@ void
 capture_tool_command(int argc, char **argv)
 {
 	/* capture tool name tool args */
-	tool_name = strrchr(argv[0], IF_WIN('\\', '/'));
+	tool_name = strrchr(argv[0], '/');
 	tool_name = tool_name ? tool_name + 1 : argv[0];
 
 	make_argv(tool_args, sizeof(tool_args), argc, argv);
@@ -2066,7 +2050,6 @@ ulonglong get_current_lsn(MYSQL *connection)
 {
 	static const char lsn_prefix[] = "\nLog sequence number ";
 	ulonglong lsn = 0;
-	msg("Getting InnoDB LSN");
 	if (MYSQL_RES *res = xb_mysql_query(connection,
 					    "SHOW ENGINE INNODB STATUS",
 					    true, false)) {
@@ -2080,10 +2063,5 @@ ulonglong get_current_lsn(MYSQL *connection)
 		}
 		mysql_free_result(res);
 	}
-	msg("InnoDB LSN: %llu, Flushing Logs", lsn);
-        /* Make sure that current LSN is written and flushed to disk. */
-	xb_mysql_query(connection, "FLUSH NO_WRITE_TO_BINLOG ENGINE LOGS",
-		       false, false);
-	msg("Flushed Logs");
 	return lsn;
 }
