@@ -285,88 +285,25 @@ int update_portion_of_time(THD *thd, TABLE *table,
   return res;
 }
 
-/**
-  Delete a record stored in:
-  replace= true: record[0]
-  replace= false: record[1]
-
-  with regard to the treat_versioned flag, which can be false for a versioned
-  table in case of versioned->versioned replication.
-
- For a versioned case, we detect a few conditions, under which we should delete
- a row instead of updating it to a history row.
- This includes:
- * History deletion by user;
- * History collision, in case of REPLACE or very fast sequence of dmls
-   so that timestamp doesn't change;
- * History collision in the parent table
-
- A normal delete is processed here as well.
-*/
-template <bool replace>
-int TABLE::delete_row(bool treat_versioned)
+inline
+int TABLE::delete_row()
 {
-  int err= 0;
-  uchar *del_buf= record[replace ? 1 : 0];
-
-  bool delete_row= !treat_versioned
-                   || in_use->lex->vers_conditions.delete_history
-                   || versioned(VERS_TRX_ID)
-                   || !vers_end_field()->is_max(
-                           vers_end_field()->ptr_in_record(del_buf));
-  if (!delete_row)
-  {
-    if (replace)
-    {
-      store_record(this, record[2]);
-      restore_record(this, record[1]);
-    }
-    else
-    {
-      store_record(this, record[1]);
-    }
-    vers_update_end();
-
-    Field *row_start= vers_start_field();
-    Field *row_end= vers_end_field();
-    // Don't make history row with negative lifetime
-    delete_row= row_start->cmp(row_start->ptr, row_end->ptr) > 0;
-
-    if (likely(!delete_row))
-      err= file->ha_update_row(record[1], record[0]);
-    if (unlikely(err))
-    {
-      /*
-        MDEV-23644: we get HA_ERR_FOREIGN_DUPLICATE_KEY iff we already got
-        history row with same trx_id which is the result of foreign key
-        action, so we don't need one more history row.
-
-        Additionally, delete the row if versioned record already exists.
-        This happens on replace, a very fast sequence of inserts and deletes,
-        or if timestamp is frozen.
-      */
-      delete_row= err == HA_ERR_FOUND_DUPP_KEY
-                  || err == HA_ERR_FOUND_DUPP_UNIQUE
-                  || err == HA_ERR_FOREIGN_DUPLICATE_KEY;
-      if (!delete_row)
-        return err;
-    }
-
-    if (delete_row)
-      del_buf= record[1];
-
-    if (replace)
-      restore_record(this, record[2]);
-  }
-
-  if (delete_row)
-    err= file->ha_delete_row(del_buf);
+  if (!versioned(VERS_TIMESTAMP) || !vers_end_field()->is_max())
+    return file->ha_delete_row(record[0]);
 
+  store_record(this, record[1]);
+  vers_update_end();
+  int err= file->ha_update_row(record[1], record[0]);
+  /*
+     MDEV-23644: we get HA_ERR_FOREIGN_DUPLICATE_KEY iff we already got history
+     row with same trx_id which is the result of foreign key action, so we
+     don't need one more history row.
+  */
+  if (err == HA_ERR_FOREIGN_DUPLICATE_KEY)
+    return file->ha_delete_row(record[0]);
   return err;
 }
 
-template int TABLE::delete_row<true>(bool treat_versioned);
-template int TABLE::delete_row<false>(bool treat_versioned);
 
 /**
   Implement DELETE SQL word.
@@ -456,18 +393,14 @@ bool mysql_delete(THD *thd, TABLE_LIST *table_list, COND *conds,
   }
 
   if (delete_history)
-  {
-    DBUG_ASSERT(conds);
     table->vers_write= false;
-  }
 
   if (returning)
     (void) result->prepare(returning->item_list, NULL);
 
-  if (!thd->lex->current_select->leaf_tables_saved)
+  if (thd->lex->current_select->first_cond_optimization)
   {
     thd->lex->current_select->save_leaf_tables(thd);
-    thd->lex->current_select->leaf_tables_saved= true;
     thd->lex->current_select->first_cond_optimization= 0;
   }
   /* check ORDER BY even if it can be ignored */
@@ -504,10 +437,10 @@ bool mysql_delete(THD *thd, TABLE_LIST *table_list, COND *conds,
     DBUG_RETURN(TRUE);
   }
 
-  const_cond_result= const_cond && (!conds || conds->val_bool());
+  const_cond_result= const_cond && (!conds || conds->val_int());
   if (unlikely(thd->is_error()))
   {
-    /* Error evaluating val_bool(). */
+    /* Error evaluating val_int(). */
     DBUG_RETURN(TRUE);
   }
 
@@ -590,13 +523,6 @@ bool mysql_delete(THD *thd, TABLE_LIST *table_list, COND *conds,
     if (thd->binlog_for_noop_dml(transactional_table))
       DBUG_RETURN(1);
 
-    if (!thd->lex->current_select->leaf_tables_saved)
-    {
-      thd->lex->current_select->save_leaf_tables(thd);
-      thd->lex->current_select->leaf_tables_saved= true;
-      thd->lex->current_select->first_cond_optimization= 0;
-    }
-
     my_ok(thd, 0);
     DBUG_RETURN(0);
   }
@@ -612,7 +538,8 @@ bool mysql_delete(THD *thd, TABLE_LIST *table_list, COND *conds,
   if (unlikely(error))
     DBUG_RETURN(TRUE);
   if ((select && select->check_quick(thd, safe_update, limit,
-                                     Item_func::BITMAP_ALL)) || !limit)
+                                     Item_func::BITMAP_ALL)) || !limit ||
+      table->stat_records() == 0)
   {
     query_plan.set_impossible_where();
     if (thd->lex->describe || thd->lex->analyze_stmt)
@@ -632,13 +559,6 @@ bool mysql_delete(THD *thd, TABLE_LIST *table_list, COND *conds,
     if (thd->binlog_for_noop_dml(transactional_table))
       DBUG_RETURN(1);
 
-    if (!thd->lex->current_select->leaf_tables_saved)
-    {
-      thd->lex->current_select->save_leaf_tables(thd);
-      thd->lex->current_select->leaf_tables_saved= true;
-      thd->lex->current_select->first_cond_optimization= 0;
-    }
-
     my_ok(thd, 0);
     DBUG_RETURN(0);				// Nothing to delete
   }
@@ -815,8 +735,8 @@ bool mysql_delete(THD *thd, TABLE_LIST *table_list, COND *conds,
     table->mark_columns_needed_for_delete();
   }
 
-  if (!table->prepare_triggers_for_delete_stmt_or_event() &&
-      table->file->ha_table_flags() & HA_CAN_FORCE_BULK_DELETE)
+  if ((table->file->ha_table_flags() & HA_CAN_FORCE_BULK_DELETE) &&
+      !table->prepare_triggers_for_delete_stmt_or_event())
     will_batch= !table->file->start_bulk_delete();
 
   /*
@@ -848,21 +768,27 @@ bool mysql_delete(THD *thd, TABLE_LIST *table_list, COND *conds,
                                              MEM_STRIP_BUF_SIZE);
 
     THD_STAGE_INFO(thd, stage_searching_rows_for_update);
-    while (!(error=info.read_record()) && !thd->killed && !thd->is_error())
+    while (!(error=info.read_record()) && !thd->killed &&
+          ! thd->is_error())
     {
       if (record_should_be_deleted(thd, table, select, explain, delete_history))
       {
         table->file->position(table->record[0]);
-        if ((error= deltempfile->unique_add((char*) table->file->ref)))
-          break;
+        if (unlikely((error=
+                      deltempfile->unique_add((char*) table->file->ref))))
+        {
+          error= 1;
+          goto terminate_delete;
+        }
         if (!--tmplimit && using_limit)
           break;
       }
     }
     end_read_record(&info);
-    if (table->file->ha_index_or_rnd_end() || error > 0 ||
-        deltempfile->get(table) ||
-        init_read_record(&info, thd, table, 0, &deltempfile->sort, 0, 1, 0))
+    if (unlikely(deltempfile->get(table)) ||
+        unlikely(table->file->ha_index_or_rnd_end()) ||
+        unlikely(init_read_record(&info, thd, table, 0, &deltempfile->sort, 0,
+                                  1, false)))
     {
       error= 1;
       goto terminate_delete;
@@ -1005,10 +931,9 @@ cleanup:
     query_cache_invalidate3(thd, table_list, 1);
   }
 
-  if (!thd->lex->current_select->leaf_tables_saved)
+  if (thd->lex->current_select->first_cond_optimization)
   {
     thd->lex->current_select->save_leaf_tables(thd);
-    thd->lex->current_select->leaf_tables_saved= true;
     thd->lex->current_select->first_cond_optimization= 0;
   }
 
@@ -1195,11 +1120,10 @@ int mysql_prepare_delete(THD *thd, TABLE_LIST *table_list, Item **conds,
 ***************************************************************************/
 
 
-extern "C" int refpos_order_cmp(void *arg, const void *a, const void *b)
+extern "C" int refpos_order_cmp(void* arg, const void *a,const void *b)
 {
-  auto file= static_cast<handler *>(arg);
-  return file->cmp_ref(static_cast<const uchar *>(a),
-                       static_cast<const uchar *>(b));
+  handler *file= (handler*)arg;
+  return file->cmp_ref((const uchar*)a, (const uchar*)b);
 }
 
 /*
@@ -1516,13 +1440,6 @@ void multi_delete::abort_result_set()
 {
   DBUG_ENTER("multi_delete::abort_result_set");
 
-  /****************************************************************************
-
-    NOTE: if you change here be aware that almost the same code is in
-     multi_delete::send_eof().
-
-  ***************************************************************************/
-
   /* the error was handled or nothing deleted and no side effects return */
   if (error_handled ||
       (!thd->transaction->stmt.modified_non_trans_table && !deleted))
@@ -1675,7 +1592,7 @@ int multi_delete::do_table_deletes(TABLE *table, SORT_INFO *sort_info,
       during ha_delete_row.
       Also, don't execute the AFTER trigger if the row operation failed.
     */
-    if (likely(!local_error))
+    if (unlikely(!local_error))
     {
       deleted++;
       if (table->triggers &&
@@ -1725,13 +1642,6 @@ bool multi_delete::send_eof()
   /* reset used flags */
   THD_STAGE_INFO(thd, stage_end);
 
-  /****************************************************************************
-
-    NOTE: if you change here be aware that almost the same code is in
-     multi_delete::abort_result_set().
-
-  ***************************************************************************/
-
   if (thd->transaction->stmt.modified_non_trans_table)
     thd->transaction->all.modified_non_trans_table= TRUE;
   thd->transaction->all.m_unsafe_rollback_flags|=
